#!/usr/bin/env python3

import sys
import os
import re
import cv2
import numpy as np
import argparse
import glob
import yaml
import scipy.io
from scipy.spatial.transform import Rotation as R

# Declare img ids to test
img_ids = [f'{n:06}' for n in range(100)]

cam_K = np.array([[665.80768,0,637.642], [0,665.80754,367.56], [0,0,1]])
cam_D = np.array([0.0,0.0,0.0,0.0,0.0])


def draw_axis(img, corners, imgpts, linesize=5, colors=[(0,0,255), (0,255,0), (255,0,0)]):
	# If only one color is specified
	if np.shape(colors) != (3,3):
		colors = [colors] * 3

	corner = tuple(corners[0].astype(np.int32).ravel())
	# GBR colorspace
	# X-Red, Y-Green, Z-Blue
	img = cv2.line(img, corner, tuple(imgpts[0].astype(np.int32).ravel()), colors[0], linesize) # X - R
	img = cv2.line(img, corner, tuple(imgpts[1].astype(np.int32).ravel()), colors[1], linesize) # Y - G
	img = cv2.line(img, corner, tuple(imgpts[2].astype(np.int32).ravel()), colors[2], linesize) # Z - B
	return img

# Helper for drawing an axis with only a pose, projects the 2D origin point first
def project_and_draw_axis(img, axis, rvecs, tvecs, mtx, dist, linesize=5, colors=[(0,0,255), (0,255,0), (255,0,0)]):
	origin_pt, _ = cv2.projectPoints(np.float32([0,0,0]), rvecs, tvecs, mtx, dist)
	imgpts, _ = cv2.projectPoints(axis, rvecs, tvecs, mtx, dist)
	return draw_axis(img, origin_pt, imgpts, linesize, colors)

def draw_bounding_cuboid(img, bbox_3d, cam_K, tvec, rvec, color = (0, 255, 0), thickness = 2):
	bbox_2d, _ = cv2.projectPoints(bbox_3d, rvec, tvec, cam_K, None)
	bbox_2d = np.squeeze(bbox_2d)

	#convert bbox to int and tuple
	bbox = np.copy(bbox_2d).astype(np.int32)
	bbox = tuple(map(tuple, bbox))

	# X Face, Red
	cv2.line(img, bbox[0], bbox[1], (0, 0, 255), thickness)
	cv2.line(img, bbox[1], bbox[3], (0, 0, 255), thickness)
	cv2.line(img, bbox[2], bbox[3], (0, 0, 255), thickness)
	cv2.line(img, bbox[0], bbox[2], (0, 0, 255), thickness)

	# Y Face - Green
	cv2.line(img, bbox[4], bbox[5], (0, 255, 0), thickness)
	cv2.line(img, bbox[0], bbox[4], (0, 255, 0), thickness)
	cv2.line(img, bbox[1], bbox[5], (0, 255, 0), thickness)

	# Z Face - Blue
	cv2.line(img, bbox[4], bbox[6], (255, 0, 0), thickness)
	cv2.line(img, bbox[2], bbox[6], (255, 0, 0), thickness)

	# Draw the rest Yellow
	cv2.line(img, bbox[5], bbox[7], (0, 255, 255), thickness)
	cv2.line(img, bbox[6], bbox[7], (0, 255, 255), thickness)
	cv2.line(img, bbox[3], bbox[7], (0, 255, 255), thickness)


def check_images(args):
	# Select cube or cracker box
	if args.object_type == 'cube':
		base_data_path = './test_data/cube'
		point_file = os.path.join(base_data_path, 'data/models/cube/points.xyz')
		xyz = np.loadtxt(point_file)
	elif args.object_type == 'cracker_box':
		base_data_path = './test_data/cracker_box'
		point_file = os.path.join(base_data_path, 'data/models/_03_cracker_box/points.xyz')
		xyz = np.loadtxt(point_file)
		# Extract extents for bounding cuboid of object
		min_x = np.min(xyz[:,0])
		max_x = np.max(xyz[:,0])
		min_y = np.min(xyz[:,1])
		max_y = np.max(xyz[:,1])
		min_z = np.min(xyz[:,2])
		max_z = np.max(xyz[:,2])
		xyz = np.array([[max_x, max_y, max_z],
						[max_x, max_y, min_z],
						[max_x, min_y, max_z],
						[max_x, min_y, min_z],
						[min_x, max_y, max_z],
						[min_x, max_y, min_z],
						[min_x, min_y, max_z],
						[min_x, min_y, min_z]])

	for img_id in img_ids:
		img_file = os.path.join(base_data_path, 'data/YCB_Video/data/0000/{}-color.png'.format(img_id))
		gt_file = os.path.join(base_data_path, 'data/YCB_Video/data/0000/{}-meta.mat'.format(img_id))
		gt = scipy.io.loadmat(gt_file)
		gt = gt['poses'].squeeze()
		tvec = gt[:,3]
		rotmat = gt[:3,:3]
		rvec = R.from_matrix(rotmat).as_rotvec()

		img = cv2.imread(img_file)

		# Project and draw cube based on vertices
		draw_bounding_cuboid(img, xyz, cam_K, tvec, rvec)

		# Project and draw center point
		center_pt_3d = np.float32([0,0,0])
		centerpt, jac = cv2.projectPoints(center_pt_3d, rvec, tvec, cam_K, cam_D)
		centerpt = np.squeeze(centerpt)
		img = cv2.circle(img, (int(centerpt[0]),int(centerpt[1])), 5, (0,0,255), 10)

		# Project and draw axis
		axis = np.float32([[0.2,0,0], [0,0.2,0], [0,0,0.2]]).reshape(-1,3)
		project_and_draw_axis(img, axis, rvec, tvec, cam_K, cam_D)

		cv2.imshow("result", img)
		key = cv2.waitKey(0) & 0xFF
		if key == 27:
			sys.exit(0)


if __name__ == "__main__":
	parser = argparse.ArgumentParser()
	parser.add_argument('object_type', choices=['cube', 'cracker_box'], help = 'Object type to test')
	args = parser.parse_args()
	
	check_images(args)
